import math
import random

import cocos
import pyglet

from cocos.rect import Rect

from gamelib.euclid import *

import curve

CURVE = math.radians(45)
MID = math.pi - CURVE*2

branch = pyglet.graphics.OrderedGroup(0)
path_background = pyglet.graphics.OrderedGroup(1)
path_foreground = pyglet.graphics.OrderedGroup(2)

class Branch(object):
    '''
    
    rect - my bounding rect to determine if I'm on-screen
    curves - my curves (p1, p2, cp1, cp2, d)

    vertex_list
    '''
    def __init__(self, batch, x, y, start, size, direction, width):
        # x,y is the start point
        # start is the initial vector
        # direction is the sign of the angle to curve in

        self.batch = batch
        self.x, self.y = x, y
        self.start = start
        self.size = size
        self.direction = direction
        self.width = width
        self.alpha = 0

        w = size
        self.num_curves = 0
        while w > 10:
            self.num_curves += 1
            w /= 1.618

        self.curves, self.points, colors = curve.curve_vertices(x, y, size,
            start, direction, width, math.pi/4, self.num_curves)

        self.vertex_list = None
        if self.batch is not None:
            # create the quads list
            N = len(self.points)/2
            c = [0x5d, 0x76, 0x1f, 255]*N
            l = [0, 0]*N
            self.vertex_list = self.batch.add(N, pyglet.gl.GL_QUADS, branch,
                ('v2f', l), ('c4B', c))

        self.branches = []

        if self.size > 100:
            l = range(len(self.curves))[1:-2]
            if l:
                for i in random.sample(l, random.randint(0, len(l))):
                    self.branch(i)

    def delete(self):
        self.vertex_list.delete()
        self.vertex_list = None
        for b in self.branches:
            b.delete()
        self.branches = []

    def branch(self, n):
        (p1, p2, cp1, cp2, sp, d, w) = self.curves[n]
        b = Branch(self.batch, sp.x, sp.y, cp1-p1, abs(d)*.9,
                -self.direction, w)
        self.branches.append(b)
        return b


    def update(self, dt):
        for b in self.branches:
            b.update(dt)
    do_nothing = update


    def appear(self):
        self.growth = 0
        self.alpha = 255
        self.update = self.grow_in

    def growth_done(self):
        if self.growth != 1:
            return False
        for b in self.branches:
            if not b.growth_done():
                return False
        return True

    def grow_in(self, dt):
        for b in self.branches:
            b.update(dt)

        if self.growth == 1:
            return

        self.growth += .5*dt
        if self.growth > 1:
            self.growth = 1

        g = math.sin(self.growth * math.pi/2)
        self.curves, self.points, colors = curve.curve_vertices(self.x,
            self.y, self.size*g, self.start, self.direction,
            self.width*g, math.pi/4*g, self.num_curves)

        N = len(self.points)/2
        self.vertex_list.vertices[:] = self.points

        if self.growth == 1:
            for b in self.branches:
                b.appear()


    def appear_old(self):
        self.update = self.fade_in
        self.delay = 255
        for b in self.branches:
            b.appear()

    def fade_in(self, dt):
        self.delay -= int(255*dt)
        if self.delay < 0:
            self.alpha = min(255, self.alpha + int(255*dt))
            if self.alpha == 255:
                self.update = self.do_nothing
        N = len(self.points)/2
        self.vertex_list.colors[3::4] = [self.alpha] * N
        for b in self.branches:
            b.update(dt)


    removed = False
    def remove(self):
        # will be called many times
        if self.removed:
            return
        self.removed = True
        self.update = self.fade_out
        for b in self.branches:
            b.remove()

    def fade_out(self, dt):
        if self.vertex_list is None:
            # just in case
            self.update = self.do_nothing
            return
        self.alpha = max(0, self.alpha - int(255*dt))
        if self.alpha == 0:
            self.update = self.do_nothing
        N = len(self.points)/2
        self.vertex_list.colors[3::4] = [self.alpha] * N
        for b in self.branches:
            b.update(dt)

class Path(object):
    def __init__(self, batch, x, y, start_vector, size, direction, width,
            type, last = None):
        # x,y is the start point
        # start is the initial vector
        # direction is the sign of the angle to curve in

        self.start_vector = start_vector.normalized()

        self.batch = batch
        self.x, self.y = x, y
        self.size = size
        self.direction = direction
        self.width = width
        self.type = type
        self.sign = [-1, 1][self.direction > 0]
        self.next = None

        self.points = []
        self.alphas = []
        self.branches = {}

        self.path, self.branch_points, self.quads, self.lines, \
            self.end_vector = curve.path_vertices(x, y, size, 
            start_vector, direction, width, width, type, math.pi/4)
        self.end_point = self.path[-1]
        self.quads = [0] * len(self.quads)
        self.lines = [0] * len(self.lines)

        if self.batch is not None:
            self.generate_lists()

        self.add_branches()

    is_alive = True

    def generate_lists(self):
        # create the quads list
        N = len(self.lines)/2
        c = [0x5b,0x78,0x11,255] * N
        self.alphas = [255]*N
        self.vertex_list1 = self.batch.add(N, pyglet.gl.GL_QUADS, path_background,
            ('v2f', self.quads), ('c4B', c))
        c = [0x75,0x8b,0x40,255] * N
        self.vertex_list2 = self.batch.add(N, pyglet.gl.GL_LINES, path_foreground,
            ('v2f', self.lines), ('c4B', c))

    appeared = False
    growth = 0
    def appear(self):
        if self.appeared:
            return

        # scale/grow instead of fading in 
        self.appeared = True
        self.update = self.grow_in

    def growth_done(self):
        if self.growth != 1:
            return False
        for i in self.branches:
            if not self.branches[i].growth_done():
                return False
        return True

    def grow_in(self, dt):
        for i in self.branches:
            self.branches[i].update(dt)

        if self.growth == 1:
            self.update = self.do_nothing
            return

        self.growth += .5*dt
        if self.growth > 1:
            self.growth = 1

        g = math.sin(self.growth * math.pi/2)
        self.path, self.branch_points, self.quads, self.lines, x = \
            curve.path_vertices(self.x, self.y, self.size*g, 
            self.start_vector, self.direction, self.width, self.width*g,
            self.type, math.pi/4*g)

        self.vertex_list1.vertices[:] = self.quads
        self.vertex_list2.vertices[:] = self.lines

        if self.growth == 1:
            for i in self.branches:
                self.branches[i].appear()


    fade_up_to = 0
    def remove_segment(self, i):
        self.update = self.fade_out
        self.fade_up_to = i
        for b in self.branches:
            if b < i:
                self.branches[b].remove()

    def update(self, dt):
        for i in self.branches:
            self.branches[i].update(dt)
    do_nothing = update

    def fade_in(self, dt):
        a = self.alphas[0]
        a = min(255, a + max(1, int(150*dt)))
        self.alphas[:] = [a] * len(self.alphas)
        l = [a] * len(self.alphas)
        l2 = [a] * len(self.alphas)
        self.vertex_list1.colors[3::4] = l
        self.vertex_list2.colors[3::4] = l2
        if a == 255:
            self.update = self.do_nothing
        for b in self.branches:
            self.branches[b].update(dt)

    def fade_out(self, dt):
        if not self.is_alive:
            return
        for b in self.branches:
            self.branches[b].update(dt)
        if not self.fade_up_to:
            return
        l = []
        l2 = []
        for i in range(self.fade_up_to+1):
            a = self.alphas[i]
            self.alphas[i] = a = max(0, a - max(1, int(150*dt)))
            if i == 0: n = 2
            elif i == self.fade_up_to: n = 2
            else: n = 4
            l.extend([a]*n)
            l2.extend([a]*n)
        if i == len(self.path)-1 and a == 0:
            for b in self.branches:
                self.branches[b].delete()
            self.delete()
            self.is_alive = False
            return
        self.vertex_list1.colors[3:len(l)*4:4] = l
        self.vertex_list2.colors[3:len(l2)*4:4] = l2

    def delete(self):
        self.vertex_list1.delete()
        self.vertex_list2.delete()
        self.vertex_list1 = self.vertex_list2 = None

    def add(self, size, direction, type):
        v = self.end_vector
        x, y = self.end_point
        self.next = Path(self.batch, x, y, -v, size,
            direction, self.width, type)
        return self.next

    def add_branches(self):
        if self.size > 200:
            l = [0,1]
        else:
            l = [random.randint(0, 1)]
        for n in l:
            (ep1, p2, ep2, v, i) = self.branch_points[n]
            if self.type == 1:
                b = Branch(self.batch, ep1.x, ep1.y, v, self.size/1.618,
                    -self.sign, self.width)
            else:
                if n:
                    b = Branch(self.batch, ep2.x, ep2.y, v, self.size/1.618,
                        self.sign, self.width)
                else:
                    b = Branch(self.batch, ep1.x, ep1.y, v, self.size/1.618,
                        -self.sign, self.width)
            self.branches[i] = b


LONG = 400
MEDIUM = 400/1.618

patterns = [
    [   # start
        (MEDIUM, -1, 1),
        (MEDIUM, 1, 1),
    ],
    [   # long
        (LONG, -1, 1),
        (LONG, 1, 1),
    ],
    [   # wriggly
        (LONG, -1, 1),
        (MEDIUM, 1, 1),
        (LONG, -1, 2),
        (MEDIUM, -1, 1),
        (MEDIUM, 1, 1),
        (LONG, 1, 1),
    ],
    [   # circle
        (LONG, -1, 1),
        (LONG, -1, 2),
        (MEDIUM, -1, 1),
        (LONG, -1, 2),
        (MEDIUM, -1, 1),
        (LONG, -1, 2),
        (LONG, -1, 1),
    ],
]

class TreeLayer(cocos.layer.ScrollableLayer):
    def __init__(self):
        super(TreeLayer, self).__init__()
        node = cocos.batch.BatchNode()
        self.add(node)
        self.batch = node.batch
        p = Path(self.batch, 0, 0, Vector2(1, 0), LONG, 1, 20, 1)
        self.path_objects = [p]
        self.path = [(p, i, x, y) for i, (x, y) in enumerate(p.path)]
        self.construct_path(-1)
        while self.to_construct:
            self.construct()
        p.appear()

    def on_enter(self):
        super(TreeLayer, self).on_exit()
        self.game = self.get_ancestor(cocos.scene.Scene)

    def on_exit(self):
        super(TreeLayer, self).on_exit()
        self.game = None

    def update(self, dt):
        if self.game.starting:
            p = self.path_objects[0]
            if p.growth_done():
                self.game.starting = False
            else:
                p.update(dt)
                p.next.update(dt)
            return
        for p in list(self.path_objects):
            p.update(dt)
            if not p.is_alive:
                self.path_objects.remove(p)
                self.path = [e if (e and e[0] != p) else None
                    for e in self.path]

        self.construct()

    def construct(self):
        if self.to_construct:
            info = self.to_construct.pop(0)
            p = self.path_objects[-1].add(*info)
            self.path_objects.append(p)
            l = [(p, i, x, y) for i, (x, y) in enumerate(p.path)]
            self.path.extend(l)

        if len(self.path_objects) + len(self.to_construct) < 6:
            self.construct_path()

    def construct_path(self, n=None):
        if n is not None:
            pattern = patterns[n]
        else:
            pattern = random.choice(patterns)
        self.to_construct = list(pattern)

    def move_along_path(self, move, position, index, player=None):
        old = x, y = position
        increment = player and 1 or -1
        while move:
            try:
                segment = self.path[index]
            except IndexError:
                return None
            if player:
                while segment is None:
                    index += increment
                    segment = self.path[index]
            elif segment is None:
                return

            path, path_index, sx, sy = segment

            if player and path is not player.current_path:
                path.next.appear()
                path.next.next.appear()

            # vector from current path position to the end of the segment
            segment = Vector2(sx-x, sy-y)
            segment_length = abs(segment)
            if move < segment_length:
                dx, dy = segment.normalized() * move
                position = (x+dx, y+dy)

        #        if player:
        #            a = 90-math.degrees(math.atan2(sy-y, sx-x))
        #            if move > 10: d = .1
        #            else: d = .3
        #            player.player.do(cocos.actions.RotateTo(a, duration=d))
                break

            if player:
                path.remove_segment(path_index)
                player.current_path = path

            # move player to the end of that segment then
            move -= segment_length
            position = sx, sy

            # next segment, and possibly next path
            index += increment

        return index, position, segment


def speed_test():
    b = Branch(None, 0, 0, Vector2(1, 0), 200, 1, 20)
    p = Path(None, 0, 0, Vector2(1, 0), 200, 1, 20, 1)
    b = Branch(None, 0, 0, Vector2(1, 0), 200, 1, 20)
    p = Path(None, 0, 0, Vector2(1, 0), 200, 1, 20, 1)

